/* USER CODE BEGIN Header */
/**
 ******************************************************************************
 * @file           : main.c
 * @brief          : Main program body
 ******************************************************************************
 * @attention
 *
 * Copyright (c) 2022 STMicroelectronics.
 * All rights reserved.
 *
 * This software is licensed under terms that can be found in the LICENSE file
 * in the root directory of this software component.
 * If no LICENSE file comes with this software, it is provided AS-IS.
 *
 ******************************************************************************
 */
/* USER CODE END Header */
/* Includes ------------------------------------------------------------------*/
#include "main.h"
#include "i2c.h"
#include "tim.h"
#include "usart.h"
#include "gpio.h"

/* Private includes ----------------------------------------------------------*/
/* USER CODE BEGIN Includes */
#include "onewire.h"
#include "ds18b20.h"
#include "string.h"
#include "ssd1306.h"
#include "ssd1306_tests.h"
#include "stdio.h"
/* USER CODE END Includes */

/* Private typedef -----------------------------------------------------------*/
/* USER CODE BEGIN PTD */

/* USER CODE END PTD */

/* Private define ------------------------------------------------------------*/
/* USER CODE BEGIN PD */
/* USER CODE END PD */

/* Private macro -------------------------------------------------------------*/
/* USER CODE BEGIN PM */

/* USER CODE END PM */

/* Private variables ---------------------------------------------------------*/

/* USER CODE BEGIN PV */
float temperature;
float average_temperature;
float previous_average_temperature;
char message[64];
char lcd_line[22];
const float temperature_threshold = 36.0;
const float temperature_change_threshold = 0.05;
/* USER CODE END PV */

/* Private function prototypes -----------------------------------------------*/
void SystemClock_Config(void);
/* USER CODE BEGIN PFP */

/* USER CODE END PFP */

/* Private user code ---------------------------------------------------------*/
/* USER CODE BEGIN 0 */

/* USER CODE END 0 */

/**
 * @brief  The application entry point.
 * @retval int
 */
int main(void)
{
	/* USER CODE BEGIN 1 */

	/* USER CODE END 1 */

	/* MCU Configuration--------------------------------------------------------*/

	/* Reset of all peripherals, Initializes the Flash interface and the Systick. */
	HAL_Init();

	/* USER CODE BEGIN Init */

	/* USER CODE END Init */

	/* Configure the system clock */
	SystemClock_Config();

	/* USER CODE BEGIN SysInit */

	/* USER CODE END SysInit */

	/* Initialize all configured peripherals */
	MX_GPIO_Init();
	MX_USART2_UART_Init();
	MX_I2C1_Init();
	MX_TIM7_Init();
	/* USER CODE BEGIN 2 */
	DS18B20_Init(DS18B20_Resolution_12bits);

	ssd1306_Init();
	ssd1306_Fill(Black);
	ssd1306_SetCursor(0, 5);
	ssd1306_WriteString("DS18B20 (x3) 1-Wire", Font_6x8, White);
	ssd1306_SetCursor(0, 25);
	ssd1306_WriteString("temperature sensors", Font_6x8, White);
	ssd1306_SetCursor(0, 45);
	ssd1306_WriteString("Library: msalamon.pl", Font_6x8, White);
	ssd1306_UpdateScreen();

	HAL_Delay(3000);

	//  HAL_GPIO_WritePin(TEST_GPIO_Port, TEST_Pin, 0);
	/* USER CODE END 2 */

	/* Infinite loop */
	/* USER CODE BEGIN WHILE */
	while (1)
	{

		DS18B20_ReadAll();
		//  HAL_GPIO_WritePin(TEST_GPIO_Port, TEST_Pin, 1);
		DS18B20_StartAll();
		//   HAL_GPIO_WritePin(TEST_GPIO_Port, TEST_Pin, 0);
		uint8_t ROM_tmp[8];
		uint8_t i;
		ssd1306_Fill(Black);
		ssd1306_SetCursor(0, 0);
		ssd1306_WriteString("Temperature readings:", Font_6x8, White);
		HAL_GPIO_WritePin(Red_LED_GPIO_Port, Red_LED_Pin, RESET);

		average_temperature = 0.0;

		for (i = 0; i < DS18B20_Quantity(); i++)
		{
			if (DS18B20_GetTemperature(i, &temperature))
			{
				DS18B20_GetROM(i, ROM_tmp);
				memset(message, 0, sizeof(message));
				sprintf(message, "%d. ROM: %X%X%X%X%X%X%X%X Temp: %f\n\r", i,
						ROM_tmp[0], ROM_tmp[1], ROM_tmp[2], ROM_tmp[3],
						ROM_tmp[4], ROM_tmp[5], ROM_tmp[6], ROM_tmp[7],
						temperature);
				HAL_UART_Transmit(&huart2, (uint8_t*) message, sizeof(message),
						100);
				ssd1306_SetCursor(0, (i * 10) + 15);
				if (temperature >= temperature_threshold)
				{
					sprintf(lcd_line, "T%d: %.2f deg. C !!!", i + 1,
							temperature);
					HAL_GPIO_WritePin(Red_LED_GPIO_Port, Red_LED_Pin, SET);
				}
				else
				{
					sprintf(lcd_line, "T%d: %.2f deg. C", i + 1, temperature);

				}
				ssd1306_WriteString(lcd_line, Font_6x8, White);
				average_temperature += temperature;

			}
		}

		average_temperature = average_temperature / DS18B20_Quantity();
		if (average_temperature
				> previous_average_temperature + temperature_change_threshold)
		{
			sprintf(lcd_line, "Ta: %.2f deg. C (+)", average_temperature);
		}
		else if (average_temperature
				< previous_average_temperature - temperature_change_threshold)
		{
			sprintf(lcd_line, "Ta: %.2f deg. C (-)", average_temperature);
		}
		else
		{
			sprintf(lcd_line, "Ta: %.2f deg. C", average_temperature);
		}

		ssd1306_SetCursor(0, 50);
		ssd1306_WriteString(lcd_line, Font_6x8, White);
		ssd1306_UpdateScreen();
		HAL_UART_Transmit(&huart2, (uint8_t*) "\n\r", sizeof("\n\r"), 100);

		HAL_GPIO_TogglePin(Green_LED_GPIO_Port, Green_LED_Pin);
		previous_average_temperature = average_temperature;
		HAL_Delay(1000);

		/* USER CODE END WHILE */

		/* USER CODE BEGIN 3 */
	}
	/* USER CODE END 3 */
}

/**
 * @brief System Clock Configuration
 * @retval None
 */
void SystemClock_Config(void)
{
	RCC_OscInitTypeDef RCC_OscInitStruct =
	{ 0 };
	RCC_ClkInitTypeDef RCC_ClkInitStruct =
	{ 0 };

	/** Configure the main internal regulator output voltage
	 */
	if (HAL_PWREx_ControlVoltageScaling(PWR_REGULATOR_VOLTAGE_SCALE1) != HAL_OK)
	{
		Error_Handler();
	}

	/** Configure LSE Drive Capability
	 */
	HAL_PWR_EnableBkUpAccess();
	__HAL_RCC_LSEDRIVE_CONFIG(RCC_LSEDRIVE_LOW);

	/** Initializes the RCC Oscillators according to the specified parameters
	 * in the RCC_OscInitTypeDef structure.
	 */
	RCC_OscInitStruct.OscillatorType = RCC_OSCILLATORTYPE_LSE
			| RCC_OSCILLATORTYPE_MSI;
	RCC_OscInitStruct.LSEState = RCC_LSE_ON;
	RCC_OscInitStruct.MSIState = RCC_MSI_ON;
	RCC_OscInitStruct.MSICalibrationValue = 0;
	RCC_OscInitStruct.MSIClockRange = RCC_MSIRANGE_6;
	RCC_OscInitStruct.PLL.PLLState = RCC_PLL_ON;
	RCC_OscInitStruct.PLL.PLLSource = RCC_PLLSOURCE_MSI;
	RCC_OscInitStruct.PLL.PLLM = 1;
	RCC_OscInitStruct.PLL.PLLN = 40;
	RCC_OscInitStruct.PLL.PLLP = RCC_PLLP_DIV7;
	RCC_OscInitStruct.PLL.PLLQ = RCC_PLLQ_DIV2;
	RCC_OscInitStruct.PLL.PLLR = RCC_PLLR_DIV2;
	if (HAL_RCC_OscConfig(&RCC_OscInitStruct) != HAL_OK)
	{
		Error_Handler();
	}

	/** Initializes the CPU, AHB and APB buses clocks
	 */
	RCC_ClkInitStruct.ClockType = RCC_CLOCKTYPE_HCLK | RCC_CLOCKTYPE_SYSCLK
			| RCC_CLOCKTYPE_PCLK1 | RCC_CLOCKTYPE_PCLK2;
	RCC_ClkInitStruct.SYSCLKSource = RCC_SYSCLKSOURCE_PLLCLK;
	RCC_ClkInitStruct.AHBCLKDivider = RCC_SYSCLK_DIV1;
	RCC_ClkInitStruct.APB1CLKDivider = RCC_HCLK_DIV1;
	RCC_ClkInitStruct.APB2CLKDivider = RCC_HCLK_DIV1;

	if (HAL_RCC_ClockConfig(&RCC_ClkInitStruct, FLASH_LATENCY_4) != HAL_OK)
	{
		Error_Handler();
	}

	/** Enable MSI Auto calibration
	 */
	HAL_RCCEx_EnableMSIPLLMode();
}

/* USER CODE BEGIN 4 */

/* USER CODE END 4 */

/**
 * @brief  This function is executed in case of error occurrence.
 * @retval None
 */
void Error_Handler(void)
{
	/* USER CODE BEGIN Error_Handler_Debug */
	/* User can add his own implementation to report the HAL error return state */
	__disable_irq();
	while (1)
	{
	}
	/* USER CODE END Error_Handler_Debug */
}

#ifdef  USE_FULL_ASSERT
/**
  * @brief  Reports the name of the source file and the source line number
  *         where the assert_param error has occurred.
  * @param  file: pointer to the source file name
  * @param  line: assert_param error line source number
  * @retval None
  */
void assert_failed(uint8_t *file, uint32_t line)
{
  /* USER CODE BEGIN 6 */
  /* User can add his own implementation to report the file name and line number,
     ex: printf("Wrong parameters value: file %s on line %d\r\n", file, line) */
  /* USER CODE END 6 */
}
#endif /* USE_FULL_ASSERT */
